@@ -15,6 +15,7 @@
 
 
 /* Command line parsing and dispatching */
+
 #include "ctype_.h"
 #include "memory_.h"
 #include "string_.h"
@@ -26,9 +27,13 @@
 #include "gscdefs.h"
 #include "gsmalloc.h"           /* for gs_malloc_limit */
 #include "gsmdebug.h"
+#include "gspaint.h"		/* for gs_erasepage */
 #include "gxdevice.h"
 #include "gxdevmem.h"
 #include "gsdevice.h"
+#include "gxdevsop.h"		/* for gxdso_* enums */
+#include "gxclpage.h"
+#include "gdevprn.h"
 #include "stream.h"
 #include "ierrors.h"
 #include "estack.h"
@@ -134,7 +139,9 @@ gs_main_init_with_args(gs_main_instance * minst, int argc, char *argv[])
     int code;
 
     arg_init(&args, (const char **)argv, argc,
-             gs_main_arg_fopen, (void *)minst);
+             gs_main_arg_fopen, (void *)minst,
+             minst->get_codepoint,
+             minst->heap);
     code = gs_main_init0(minst, 0, 0, 0, GS_MAX_LIB_DIRS);
     if (code < 0)
         return code;
@@ -167,18 +174,18 @@ gs_main_init_with_args(gs_main_instance * minst, int argc, char *argv[])
         bool helping = false;
 
         for (i = 1; i < argc; ++i)
-            if (!strcmp(argv[i], "--")) {
+            if (!arg_strcmp(&args, argv[i], "--")) {
                 /* A PostScript program will be interpreting all the */
                 /* remaining switches, so stop scanning. */
                 helping = false;
                 break;
-            } else if (!strcmp(argv[i], "--help")) {
+            } else if (!arg_strcmp(&args, argv[i], "--help")) {
                 print_help(minst);
                 helping = true;
-            } else if (!strcmp(argv[i], "--debug")) {
+            } else if (!arg_strcmp(&args, argv[i], "--debug")) {
                 gs_debug_flags_list(minst->heap);
                 helping = true;
-            } else if (!strcmp(argv[i], "--version")) {
+            } else if (!arg_strcmp(&args, argv[i], "--version")) {
                 print_version(minst);
                 puts(minst->heap, "");  /* \n */
                 helping = true;
@@ -201,11 +208,16 @@ gs_main_init_with_args(gs_main_instance * minst, int argc, char *argv[])
             (char *)gs_alloc_bytes(minst->heap, len, "GS_OPTIONS");
 
             gp_getenv(GS_OPTIONS, opts, &len);  /* can't fail */
+#if defined(GS_NO_UTF8)
             if (arg_push_memory_string(&args, opts, false, minst->heap))
                 return e_Fatal;
+#else
+            if (arg_push_decoded_memory_string(&args, opts, false, true, minst->heap))
+                return e_Fatal;
+#endif
         }
     }
-    while ((arg = arg_next(&args, &code)) != 0) {
+    while ((arg = arg_next(&args, &code, minst->heap)) != 0) {
         switch (*arg) {
             case '-':
                 code = swproc(minst, arg, &args);
@@ -216,16 +228,29 @@ gs_main_init_with_args(gs_main_instance * minst, int argc, char *argv[])
                 if (gs_debug[':'] && arg[1] == 'Z') {
                     int i;
 
-                    dprintf1("%% Init started, instance 0x%p, with args: ", minst);
+                    dmprintf1(minst->heap, "%% Init started, instance 0x%p, with args: ", minst);
                     for (i=1; i<argc; i++)
-                        dprintf1("%s ", argv[i]);
-                    dprintf("\n");
+                        dmprintf1(minst->heap, "%s ", argv[i]);
+                    dmprintf(minst->heap, "\n");
                 }
                 break;
             default:
+                /* default is to treat this as a file name to be run */
                 code = argproc(minst, arg);
                 if (code < 0)
                     return code;
+                if (minst->saved_pages_test_mode) {
+                    gx_device *pdev;
+
+                    /* get the current device */
+                    pdev = gs_currentdevice(minst->i_ctx_p->pgs);
+                    if ((code = gx_saved_pages_param_process((gx_device_printer *)pdev,
+                               (byte *)"print normal flush", 18)) < 0)
+                        return code;
+                    if (code > 0)
+                        if ((code = gs_erasepage(minst->i_ctx_p->pgs)) < 0)
+                            return code;
+                }
         }
     }
     if (code < 0)
@@ -238,16 +263,30 @@ gs_main_init_with_args(gs_main_instance * minst, int argc, char *argv[])
     if (gs_debug[':']) {
         int i;
 
-        dprintf1("%% Init done, instance 0x%p, with args: ", minst);
+        dmprintf1(minst->heap, "%% Init done, instance 0x%p, with args: ", minst);
         for (i=1; i<argc; i++)
-            dprintf1("%s ", argv[i]);
-        dprintf("\n");
+            dmprintf1(minst->heap, "%s ", argv[i]);
+        dmprintf(minst->heap, "\n");
     }
     if (!minst->run_start)
         return e_Quit;
     return code ;
 }
 
+/*
+ * Specify a decoding function
+ */
+void gs_main_inst_arg_decode(gs_main_instance * minst,
+                             gs_arg_get_codepoint *get_codepoint)
+{
+    minst->get_codepoint = get_codepoint;
+}
+
+gs_arg_get_codepoint *gs_main_inst_get_arg_decode(gs_main_instance * minst)
+{
+    return minst->get_codepoint;
+}
+
 /*
  * Run the 'start' procedure (after processing the command line).
  */
@@ -293,6 +332,18 @@ run_stdin:
             code = run_string(minst, ".runstdin", runFlush);
             if (code < 0)
                 return code;
+            if (minst->saved_pages_test_mode) {
+                gx_device *pdev;
+
+                /* get the current device */
+                pdev = gs_currentdevice(minst->i_ctx_p->pgs);
+                if ((code = gx_saved_pages_param_process((gx_device_printer *)pdev,
+                           (byte *)"print normal flush", 18)) < 0)
+                    return code;
+                if (code > 0)
+                    if ((code = gs_erasepage(minst->i_ctx_p->pgs)) < 0)
+                        return code;
+            }
             break;
         case '-':               /* run with command line args */
             if (strncmp(arg, "debug=", 6) == 0) {
@@ -300,6 +351,42 @@ run_stdin:
                 if (code < 0)
                     return code;
                 break;
+            } else if (strncmp(arg, "saved-pages=", 12) == 0) {
+                gx_device *pdev;
+                gx_device_printer *ppdev;
+
+                /* If init2 not yet done, just save the argument for processing then */
+                if (minst->init_done < 2) {
+                    if (minst->saved_pages_initial_arg == NULL) {
+                        /* Tuck the parameters away for later when init2 is done */
+                        minst->saved_pages_initial_arg = (char *)arg+12;
+                    } else {
+                        outprintf(minst->heap,
+                                  "   Only one --saved-pages=... command allowed before processing input\n");
+                        arg_finit(pal);
+                        return e_Fatal;
+                    }
+                } else {
+                    /* get the current device */
+                    pdev = gs_currentdevice(minst->i_ctx_p->pgs);
+                    if (dev_proc(pdev, dev_spec_op)(pdev, gxdso_supports_saved_pages, NULL, 0) == 0) {
+                        outprintf(minst->heap,
+                                  "   --saved-pages not supported by the '%s' device.\n",
+                                  pdev->dname);
+                        arg_finit(pal);
+                        return e_Fatal;
+                    }
+                    ppdev = (gx_device_printer *)pdev;
+                    code = gx_saved_pages_param_process(ppdev, (byte *)arg+12, strlen(arg+12));
+                    if (code > 0)
+                        if ((code = gs_erasepage(minst->i_ctx_p->pgs)) < 0)
+                            return code;
+                }
+                break;
+            /* The following code is only to allow regression testing of saved-pages */
+            } else if (strncmp(arg, "saved-pages-test", 16) == 0) {
+                minst->saved_pages_test_mode = true;
+                break;
             }
             /* FALLTHROUGH */
         case '+':
@@ -307,7 +394,7 @@ run_stdin:
             /* FALLTHROUGH */
         case '@':               /* ditto with @-expansion */
             {
-                const char *psarg = arg_next(pal, &code);
+                const char *psarg = arg_next(pal, &code, minst->heap);
 
                 if (code < 0)
                     return e_Fatal;
@@ -324,14 +411,14 @@ run_stdin:
                 if (code >= 0)
                     code = run_string(minst, "userdict/ARGUMENTS[", 0);
                 if (code >= 0)
-                    while ((arg = arg_next(pal, &code)) != 0) {
+                    while ((arg = arg_next(pal, &code, minst->heap)) != 0) {
                         code = runarg(minst, "", arg, "", runInit);
                         if (code < 0)
                             break;
                     }
                 if (code >= 0)
                     code = runarg(minst, "]put", psarg, ".runfile", runInit | runFlush);
-                arg_free(psarg, minst->heap);
+                arg_free((char *)psarg, minst->heap);
                 if (code >= 0)
                     code = e_Quit;
                 
@@ -375,7 +462,7 @@ run_stdin:
                 if (code < 0)
                     return code;
                 pal->expand_ats = false;
-                while ((arg = arg_next(pal, &code)) != 0) {
+                while ((arg = arg_next(pal, &code, minst->heap)) != 0) {
                     if (arg[0] == '@' ||
                         (arg[0] == '-' && !isdigit((unsigned char)arg[1]))
                         )
@@ -413,6 +500,19 @@ run_stdin:
                 code = argproc(minst, arg);
                 if (code < 0)
                     return code;
+                if (minst->saved_pages_test_mode) {
+                    gx_device *pdev;
+
+                    /* get the current device */
+                    pdev = gs_currentdevice(minst->i_ctx_p->pgs);
+                        return code;
+                    if ((code = gx_saved_pages_param_process((gx_device_printer *)pdev,
+                               (byte *)"print normal flush", 18)) < 0)
+                        return code;
+                    if (code > 0)
+                        if ((code = gs_erasepage(minst->i_ctx_p->pgs)) < 0)
+                            return code;
+                }
             }
             break;
         case 'F':               /* run file with buffer_size = 1 */
@@ -427,6 +527,18 @@ run_stdin:
                 minst->run_buffer_size = bsize;
                 if (code < 0)
                     return code;
+                if (minst->saved_pages_test_mode) {
+                    gx_device *pdev;
+
+                    /* get the current device */
+                    pdev = gs_currentdevice(minst->i_ctx_p->pgs);
+                    if ((code = gx_saved_pages_param_process((gx_device_printer *)pdev,
+                               (byte *)"print normal flush", 18)) < 0)
+                        return code;
+                    if (code > 0)
+                        if ((code = gs_erasepage(minst->i_ctx_p->pgs)) < 0)
+                            return code;
+                }
             }
             break;
         case 'g':               /* define device geometry */
@@ -456,7 +568,7 @@ run_stdin:
                 const char *path;
 
                 if (arg[0] == 0) {
-                    path = arg_next(pal, &code);
+                    path = arg_next(pal, &code, minst->heap);
                     if (code < 0)
                         return code;
                 } else
@@ -519,7 +631,7 @@ run_stdin:
                 int len;
 
                 if (arg[0] == 0) {
-                    adef = arg_next(pal, &code);
+                    adef = arg_next(pal, &code, minst->heap);
                     if (code < 0)
                         return code;
                 } else
@@ -528,6 +640,8 @@ run_stdin:
                     return code;
                 len = strlen(adef);
                 str = (char *)gs_alloc_bytes(minst->heap, (uint)len, "-o");
+                if (str == NULL)
+                    return e_VMerror;
                 memcpy(str, adef, len);
                 make_const_string(&value, a_readonly | avm_foreign,
                                   len, (const byte *)str);
@@ -594,7 +708,7 @@ run_stdin:
                 if ((code = gs_main_init1(minst)) < 0)
                     return code;
                 if (eqp == adef) {
-                    puts(minst->heap, "Usage: -dname, -dname=token, -sname=string");
+                    puts(minst->heap, "Usage: -dNAME, -dNAME=TOKEN, -sNAME=STRING");
                     return e_Fatal;
                 }
                 if (eqp == NULL) {
@@ -610,38 +724,69 @@ run_stdin:
                     *eqp++ = 0;
                     ialloc_set_space(idmemory, avm_system);
                     if (isd) {
-                        stream astream;
-                        scanner_state state;
+                        int num, i;
 
-                        s_init(&astream, NULL);
-                        sread_string(&astream,
-                                     (const byte *)eqp, strlen(eqp));
-                        gs_scanner_init_stream(&state, &astream);
-                        code = gs_scan_token(minst->i_ctx_p, &value, &state);
-                        if (code) {
-                            puts(minst->heap, "-dname= must be followed by a valid token");
-                            return e_Fatal;
-                        }
-                        if (r_has_type_attrs(&value, t_name,
-                                             a_executable)) {
-                            ref nsref;
+                        /* Check for numbers so we can provide for suffix scalers */
+                        /* Note the check for '#' is for PS "radix" numbers such as 16#ff */
+                        /* and check for '.' and 'e' or 'E' which are 'real' numbers */
+                        if ((strchr(eqp, '#') == NULL) && (strchr(eqp, '.') == NULL) &&
+                            (strchr(eqp, 'e') == NULL) && (strchr(eqp, 'E') == NULL) &&
+                            ((i = sscanf((const char *)eqp, "%d", &num)) == 1)) {
+                            char suffix = eqp[strlen(eqp) - 1];
 
-                            name_string_ref(minst->heap, &value, &nsref);
-#define string_is(nsref, str, len)\
-  (r_size(&(nsref)) == (len) &&\
-   !strncmp((const char *)(nsref).value.const_bytes, str, (len)))
-                            if (string_is(nsref, "null", 4))
-                                make_null(&value);
-                            else if (string_is(nsref, "true", 4))
-                                make_true(&value);
-                            else if (string_is(nsref, "false", 5))
-                                make_false(&value);
-                            else {
-                                puts(minst->heap,
-                                     "-dvar=name requires name=null, true, or false");
+                            switch (suffix) {
+                                case 'k':
+                                case 'K':
+                                    num *= 1024;
+                                    break;
+                                case 'm':
+                                case 'M':
+                                    num *= 1024 * 1024;
+                                    break;
+                                case 'g':
+                                case 'G':
+                                    /* caveat emptor: more than 2g will overflow */
+                                    /* and really should produce a 'real', so don't do this */
+                                    num *= 1024 * 1024 * 1024;
+                                    break;
+                                default:
+                                    break;   /* not a valid suffix or last char was digit */
+                            }
+                            make_int(&value, num);
+                        } else {
+                            /* use the PS scanner to capture other valid token types */
+                            stream astream;
+                            scanner_state state;
+
+                            s_init(&astream, NULL);
+                            sread_string(&astream,
+                                         (const byte *)eqp, strlen(eqp));
+                            gs_scanner_init_stream(&state, &astream);
+                            code = gs_scan_token(minst->i_ctx_p, &value, &state);
+                            if (code) {
+                                outprintf(minst->heap, "Invalid value for option -d%s, -dNAME= must be followed by a valid token\n", arg);
                                 return e_Fatal;
                             }
-#undef name_is_string
+                            if (r_has_type_attrs(&value, t_name,
+                                                 a_executable)) {
+                                ref nsref;
+
+                                name_string_ref(minst->heap, &value, &nsref);
+#undef string_is
+#define string_is(nsref, str, len)\
+       (r_size(&(nsref)) == (len) &&\
+       !strncmp((const char *)(nsref).value.const_bytes, str, (len)))
+                                if (string_is(nsref, "null", 4))
+                                    make_null(&value);
+                                else if (string_is(nsref, "true", 4))
+                                    make_true(&value);
+                                else if (string_is(nsref, "false", 5))
+                                    make_false(&value);
+                                else {
+                                    outprintf(minst->heap, "Invalid value for option -d%s, use -sNAME= to define string constants\n", arg);
+                                    return e_Fatal;
+                                }
+                            }
                         }
                     } else {
                         int len = strlen(eqp);
@@ -928,6 +1073,11 @@ static const char help_devices[] = "Available devices:";
 static const char help_default_device[] = "Default output device:";
 static const char help_emulators[] = "Input formats:";
 static const char help_paths[] = "Search path:";
+#ifdef HAVE_FONTCONFIG
+static const char help_fontconfig[] = "Ghostscript is also using fontconfig to search for font files\n";
+#else
+static const char help_fontconfig[] = "";
+#endif
 
 extern_gx_io_device_table();
 
@@ -1048,17 +1198,10 @@ print_emulators(const gs_main_instance *minst)
 {
     outprintf(minst->heap, "%s", help_emulators);
     {
-        const ref *pes;
+        const byte *s;
 
-        for (pes = gs_emulator_name_array;
-             pes->value.const_bytes != 0; pes++
-            )
-            /*
-             * Even though gs_emulator_name_array is declared and used as
-             * an array of string refs, each string is actually a
-             * (null terminated) C string.
-             */
-            outprintf(minst->heap, " %s", (const char *)pes->value.const_bytes);
+        for (s = gs_emulators; s[0] != 0; s += strlen((const char *)s) + 1)
+            outprintf(minst->heap, " %s", s);
     }
     outprintf(minst->heap, "\n");
 }
@@ -1104,6 +1247,7 @@ print_paths(gs_main_instance * minst)
         }
     }
     outprintf(minst->heap, "\n");
+    outprintf(minst->heap, "%s", help_fontconfig);
 }
 
 /* Print the help trailer. */
